#include <stdio.h>
#include <string.h>
#include <gmp2/gmp.h>
#include <math.h>

FILE *input, *output;

void usage ()
{
  fprintf(stderr,"Usage:  modcomp option sourcefile destfile\n"
	  "option:\n"
	  "    -c   compress\n"
	  "    -d   decompress\n"
	  "sourcefile: name of source file\n"
	  "destfile: name of destination file\n");
  exit(1);
};

void error (const char *const msg)
{
  fprintf(stderr,"%s\n",msg);
  exit(1);
};

void compress()
{
  bool used[256];
  int table[256], c, i, usedcount;
  long filelength;
  mpz_t block;

  printf("compressing\n");

  printf("counting occurencies of bytes\n");

  // reading whole file and marking used bytes
  for (i=0; i<256; i++) used[i]=false;
  while ((c=getc(input))!=EOF) used[c]=true;
  if ((filelength = ftell (input))==-1) error("error getting filesize");
  printf("filelength %ld\n",filelength);
  if (!filelength) return; // empty file coded as empty file :-)
  rewind (input);

  // calculate table and usedcount
  usedcount=0;
  for (i=0; i<256; i++) if (used[i]) table[i]=usedcount++;
  if (putc(usedcount,output)==EOF) error("error writing usedcount"); // 256 -> 0 !
  for (i=0; i<256; i++) if ((used[i])&&(putc(i,output)==EOF)) error("error writing table");
  if (fwrite(&filelength,sizeof(filelength),1,output)!=1) error("error writing filelength");

  printf("we have %d used bytes\n", (int)usedcount);
  //  for (i=0; i<256; i++) if (used[i]) printf("%d '%c' = %d\n",i,(char)i,table[i]);

  float ratio;
  ratio=log(float(usedcount));
  ratio/=log(2);
  ratio/=8;
  ratio*=100;
  printf("Compression ratio will be %f%%\n",ratio);

  mpz_init_set_ui(block,0);
  i=0;
  while ((c=getc(input))!=EOF) {
    mpz_mul_ui(block,block,usedcount);
    mpz_add_ui(block,block,table[c]);
    filelength--; if (++i==10000) { i=0; printf("todo: %ld bytes       \r",filelength);fflush(stdout); };
  };
  printf("\nWriting output\n");
  if (!mpz_out_raw(output,block)) error("error writing block");
  //  if (!mpz_out_str(stdout,10,block)) error("error writing block to stdout"); printf("\n");
  mpz_clear(block);
};

void decompress()
{
  printf("decompressing\n");
  int table[256], /*c,*/ i, usedcount;
  long filelength;
  mpz_t block, remainder;

  if ((usedcount=getc(input))==EOF) return; // empty file coded as empty file ;-)
  if (!usedcount) usedcount=256; // 0 -> 256
  for (i=0;i<usedcount;i++) if ((table[i]=getc(input))==EOF) error("error reading table");
  if (fread(&filelength,sizeof(filelength),1,input)!=1) error("error reading filelength");

  printf("filelength %ld\n",filelength);
  printf("we have %d used bytes\n",usedcount);
  //  for (i=0; i<usedcount; i++) printf("%d = %d '%c'\n",i,table[i],(char)table[i]);
  float ratio;
  ratio=log(float(usedcount));
  ratio/=log(2);
  ratio/=8;
  ratio*=100;
  printf("Compression ratio is %f%%\n",ratio);

  mpz_init(block);
  mpz_init(remainder);
  printf("reading input\n");
  if (!mpz_inp_raw(block,input)) error("error reading block");
  i=0;
  while (filelength--) {
    mpz_tdiv_qr_ui(block,remainder,block,usedcount);
    fseek(output,filelength,SEEK_SET);
    if (putc(table[mpz_get_ui(remainder)],output)==EOF) error("error writing output byte");
    if (++i==10000) { i=0; printf("todo: %ld bytes       \r",filelength);fflush(stdout); };
  };
  printf("\n");
  mpz_clear(remainder);
  mpz_clear(block);
};

int main (char argc, const char *argv[])
{
  bool shallcompress=false;

  if (sizeof(unsigned long int)!=4) error("programm is only written for 32 bit gmp library");

  if (argc!=4) usage();

  if (!strcmp(argv[1],"-c")) shallcompress = true;
  else if (!strcmp(argv[1],"-d"));
  else usage();

  if (!(input = fopen (argv[2],"rb"))) error("error opening sourcefile");
  if (!(output = fopen (argv[3],"wb"))) error("error opening destfile");

  if (shallcompress) compress(); else decompress();

  if (fclose(output)) error("error closing destfile");
  if (fclose(input)) error("error closing sourcefile");

  printf("done. no errors.\n");

  return 0;
};
